@@ -40,15 +40,14 @@ import logging
 import subprocess
 import functools
 import pathlib
+import queue
 import multiprocessing.connection
 from contextlib import contextmanager
 
 import msgpack
 import pylibmc
 
-VERSION = "0.7.1"
-LOCK_SH = 0
-LOCK_EX = 1
+VERSION = "0.8.2"
 RUN_PATH = "/run/starrydb/"
 SERVER_ADDRESS = f"{RUN_PATH}address"
 CACHE_SOCKET = f"{RUN_PATH}cache.sock"
@@ -59,17 +58,6 @@ class StarryError(Exception):
     def __init__(self, output):
         self.output = output
 
-def singleton(cls):
-    instances = dict()
-
-    @functools.wraps(cls)
-    def _singleton(*args, **kwargs):
-        if cls not in instances:
-            instances[cls] = cls(*args, **kwargs)
-        return instances[cls]
-    return _singleton
-
-@singleton
 class StarryLog(object):
     def __init__(self, file=None, lvl=logging.NOTSET):
         self.logger = None
@@ -196,6 +184,7 @@ class LibMemcachedClient(pylibmc.ThreadMappedPool):
 class StarryServer(object):
     def __init__(self, address=SERVER_ADDRESS, depots=[], cache_size=0, log_file=None, log_level="WARNING", rw_ratio=5):
         self.address = address
+        self.conn_lock = threading.RLock()
         self.cursor = 0
         self.depots = {}
         self.logger = StarryLog(log_file, getattr(logging, log_level, logging.WARNING))
@@ -225,7 +214,7 @@ class StarryServer(object):
             if "path" in depot:
                 depot_path = os.path.abspath(depot["path"])
 
-                self.depots[depot_path] = StarryDepot(depot_path, self.logger, self.rw_ratio, self.cursor, self.memcached)
+                self.depots[depot_path] = StarryDepot(depot_path, self.rw_ratio, self.cursor, self.memcached)
 
                 if depot.get("backup", False) is True:
                     self.depots[depot_path].enableBackup()
@@ -239,7 +228,8 @@ class StarryServer(object):
                 self.cursor += 1
 
     def __del__(self):
-        self.depots.clear()
+        with self.conn_lock:
+            self.depots.clear()
 
         if self.memcached is not None:
             try:
@@ -252,6 +242,49 @@ class StarryServer(object):
             except:
                 pass
 
+    def connect(self, conn):
+        depot = None
+
+        try:
+            while True:
+                (func, *param) = conn.recv()
+
+                if func == "open":
+                    (func, depot) = (func, *param)
+
+                    if depot not in self.depots:
+                        with self.conn_lock:
+                            if depot not in self.depots:
+                                self.depots[depot] = StarryDepot(depot, self.rw_ratio, self.cursor, self.memcached)
+                                self.cursor += 1
+                else:
+                    (ret, val) = (True, "")
+
+                    try:
+                        (ret, val) = getattr(self.depots[depot], func)(*param)
+                    except StarryError as e:
+                        (ret, val) = (False, e.output)
+                    except Exception as e:
+                        (ret, val) = (False, f"Caught exception: {e.__doc__}({e})")
+                    finally:
+                        if ret is False:
+                            self.logger.error(val)
+
+                        try:
+                            conn.send((ret, val))
+                        except:
+                            pass
+        except StarryError as e:
+            self.logger.error(e.output)
+            raise StarryError(e.output)
+        except:
+            pass
+        finally:
+            try:
+                conn.close()
+            except:
+                pass
+
     def start(self, backlog=1024, authkey=b'starry'):
         if pathlib.Path(str(self.address)).is_socket():
             try:
@@ -274,14 +307,7 @@ class StarryServer(object):
                     continue
 
                 try:
-                    (depot, mode) = conn.recv()
-
-                    if depot not in self.depots:
-                        self.depots[depot] = StarryDepot(depot, self.logger, self.rw_ratio, self.cursor, self.memcached)
-
-                        self.cursor += 1
-
-                    threading.Thread(target=self.depots[depot].executeCommand, args=(conn, mode)).start()
+                    threading.Thread(target=self.connect, args=(conn, )).start()
                 except Exception as e:
                     self.logger.error(f"Caught exception: {e.__doc__}({e})")
 
@@ -291,9 +317,8 @@ class StarryServer(object):
                         pass
 
 class StarryDepot(object):
-    def __init__(self, path, logger, rw_ratio, index, memcached):
+    def __init__(self, path, rw_ratio, index, memcached):
         self.database = pathlib.Path(path)
-        self.logger = logger
         self.rw_ratio = rw_ratio
         self.index = index
         self.memcached = memcached
@@ -301,8 +326,6 @@ class StarryDepot(object):
         self.fsync = False
         self.bak = False
         self.depot_lock = threading.Condition()
-        self.thread_lock = threading.Condition()
-        self.thread_blocking = False
         self.write_wait = 0
         self.read_wait = 0
         self.write_queue = []
@@ -311,62 +334,6 @@ class StarryDepot(object):
         if not self.database.is_dir():
             raise StarryError(f"path {path} doesn't exist.")
 
-    def executeCommand(self, conn, mode=LOCK_SH):
-        def _threadPredicate():
-            return not self.thread_blocking
-
-        def _depotPredicate():
-            if self.read_wait > 0 or self.write_wait > 0 or len(self.read_queue) > 0 or len(self.write_queue) > 0:
-                return False
-
-            return True
-
-        if mode == LOCK_EX:
-            with self.thread_lock:
-                self.thread_lock.wait_for(_threadPredicate)
-                self.thread_blocking = True
-
-        while True:
-            try:
-                (func, *param) = conn.recv()
-            except:
-                break
-
-            if mode == LOCK_SH:
-                with self.thread_lock:
-                    self.thread_lock.wait_for(_threadPredicate)
-            else:
-                with self.depot_lock:
-                    self.depot_lock.wait_for(_depotPredicate)
-
-            try:
-                conn.send(getattr(self, func)(*param))
-            except Exception as e:
-                self.logger.debug(f"Caught exception: {e.__doc__}({e})")
-
-                try:
-                    conn.send((False, f"Caught exception: {e.__doc__}({e})"))
-                except:
-                    pass
-
-                break
-
-        if mode == LOCK_EX:
-            with self.thread_lock:
-                self.thread_blocking = False
-                self.thread_lock.notify_all()
-
-        if self.cache is not None:
-            try:
-                self.cache.relinquish()
-            except:
-                pass
-
-        try:
-            conn.close()
-        except:
-            pass
-
     def enableFsync(self):
         self.fsync = True
 
@@ -604,7 +571,6 @@ class StarryDepot(object):
         path = pathlib.Path(path)
 
         if not path.is_dir():
-            self.logger.error(f"path {str(path)} doesn't exist.")
             return (False, f"path {str(path)} doesn't exist.")
 
         for elem in path.glob("*"):
@@ -628,7 +594,6 @@ class StarryDepot(object):
         path = pathlib.Path(path)
 
         if not path.is_dir():
-            self.logger.error(f"No {str(path)}.")
             return (False, f"No {str(path)}.")
 
         mark = {"mode": "w", "members": [None]}
@@ -655,13 +620,11 @@ class StarryDepot(object):
                     try:
                         elem.unlink()
                     except Exception as e:
-                        self.logger.error(f"Caught exception: {e.__doc__}({e})")
                         return (False, f"Caught exception: {e.__doc__}({e})")
             elif elem.name == ".backup" and elem.isdir():
                 try:
                     shutil.rmtree(elem)
                 except Exception as e:
-                    self.logger.error(f"Caught exception: {e.__doc__}({e})")
                     return (False, f"Caught exception: {e.__doc__}({e})")
 
         self.write_queue.clear()
@@ -1132,13 +1095,96 @@ class StarryDepot(object):
             self.depot_lock.notify_all()
 
 class StarryClient(object):
-    def __init__(self, path, mode=LOCK_SH, address=SERVER_ADDRESS):
-        self.client = multiprocessing.connection.Client(address, authkey=b'starry')
-        self.client.send((os.path.abspath(path), mode))
+    class Connection(object):
+        def __init__(self, conn):
+            self.conn = conn
+
+        def _callableForward(self, func, *args):
+            try:
+                self.conn.send((func, *args))
+
+                if self.conn.poll(30):
+                    return self.conn.recv()
+                else:
+                    return (False, "timeout.")
+            except Exception as e:
+                return (False, f"Caught exception: {e.__doc__}({e})")
+
+        def get(self, key):
+            return self._callableForward("get", key)
+
+        def update(self, key, value):
+            return self._callableForward("update", key, value)
+
+        def insert(self, key, value):
+            return self._callableForward("insert", key, value)
+
+        def delete(self, key):
+            return self._callableForward("delete", key)
+
+        def copy(self, old_key, new_key):
+            return self._callableForward("copy", old_key, new_key)
+
+        def rename(self, old_key, new_key):
+            return self._callableForward("rename", old_key, new_key)
+
+        def exist(self, key):
+            return self._callableForward("exist", key)
+
+        def list(self):
+            return self._callableForward("list")
+
+        def getEpoch(self):
+            return self._callableForward("getEpoch")
+
+        def backup(self):
+            return self._callableForward("backup")
+
+        def recover(self):
+            return self._callableForward("recover")
+
+        def clone(self, path):
+            return self._callableForward("clone", path)
+
+        def upgrade(self, path):
+            return self._callableForward("upgrade", path)
+
+        def clear(self):
+            return self._callableForward("clear")
+
+        def getMulti(self, keys):
+            return self._callableForward("getMulti", keys)
+
+        def getAll(self):
+            return self._callableForward("getAll")
+
+        def updateMulti(self, values):
+            return self._callableForward("updateMulti", values)
+
+        def updateAll(self, values):
+            return self._callableForward("updateAll", values)
+
+        def insertMulti(self, values):
+            return self._callableForward("insertMulti", values)
+
+        def deleteMulti(self, keys):
+            return self._callableForward("deleteMulti", keys)
+
+    def __init__(self, address=SERVER_ADDRESS, workers=1, authkey=b'starry'):
+        self.workers = queue.SimpleQueue()
+
+        for _ in range(int(workers)):
+            self.workers.put(multiprocessing.connection.Client(address, authkey=authkey))
 
     def __del__(self):
         try:
-            self.client.close()
+            while True:
+                client = self.workers.get(False)
+
+                try:
+                    client.close()
+                except:
+                    pass
         except:
             pass
 
@@ -1148,85 +1194,38 @@ class StarryClient(object):
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.__del__()
 
+    @contextmanager
+    def open(self, path, timeout=30):
+        try:
+            client = self.workers.get(timeout=timeout)
+        except queue.Empty:
+            raise StarryError("no client is available.")
+        except Exception as e:
+            raise StarryError(f"Caught exception: {e.__doc__}({e})")
+
+        try:
+            path = os.path.abspath(path)
+            client.send(("open", path))
+        except Exception as e:
+            try:
+                self.workers.put(client)
+            except:
+                pass
+
+            raise StarryError(f"Caught exception: {e.__doc__}({e})")
+
+        try:
+            yield self.Connection(client)
+        except Exception as e:
+            raise StarryError(f"Caught exception: {e.__doc__}({e})")
+        finally:
+            try:
+                self.workers.put(client)
+            except:
+                pass
+
     def close(self):
-        try:
-            self.client.close()
-        except:
-            pass
-
-    def _callableForward(self, func, *args):
-        try:
-            self.client.send((func, *args))
-        except Exception as e:
-            return (False, f"Caught exception: {e.__doc__}({e})")
-
-        try:
-            if self.client.poll(60):
-                return self.client.recv()
-            else:
-                return (False, "timeout")
-        except Exception as e:
-            return (False, f"Caught exception: {e.__doc__}({e})")
-
-    def get(self, key):
-        return self._callableForward("get", key)
-
-    def update(self, key, value):
-        return self._callableForward("update", key, value)
-
-    def insert(self, key, value):
-        return self._callableForward("insert", key, value)
-
-    def delete(self, key):
-        return self._callableForward("delete", key)
-
-    def copy(self, old_key, new_key):
-        return self._callableForward("copy", old_key, new_key)
-
-    def rename(self, old_key, new_key):
-        return self._callableForward("rename", old_key, new_key)
-
-    def exist(self, key):
-        return self._callableForward("exist", key)
-
-    def list(self):
-        return self._callableForward("list")
-
-    def getEpoch(self):
-        return self._callableForward("getEpoch")
-
-    def backup(self):
-        return self._callableForward("backup")
-
-    def recover(self):
-        return self._callableForward("recover")
-
-    def clone(self, path):
-        return self._callableForward("clone", path)
-
-    def upgrade(self, path):
-        return self._callableForward("upgrade", path)
-
-    def clear(self):
-        return self._callableForward("clear")
-
-    def getMulti(self, keys):
-        return self._callableForward("getMulti", keys)
-
-    def getAll(self):
-        return self._callableForward("getAll")
-
-    def updateMulti(self, values):
-        return self._callableForward("updateMulti", values)
-
-    def updateAll(self, values):
-        return self._callableForward("updateAll", values)
-
-    def insertMulti(self, values):
-        return self._callableForward("insertMulti", values)
-
-    def deleteMulti(self, keys):
-        return self._callableForward("deleteMulti", keys)
+        self.__del__()
 
 def _getPart(value, branch, filters):
     def _filterValue(value, filters):
